import abc
import logging
import asyncio
from typing import Any, Tuple, Optional, Iterable, AsyncIterable

from .items import Paper


logger = logging.getLogger("graph")


class Crawler(metaclass=abc.ABCMeta):
    def __init__(self, paperId_list: list[str]) -> None:
        self.papers: dict[str, Paper] = {paperId: None for paperId in paperId_list}
        self.checked = set()
        self.ref_idx: dict[str, set[str]] = {}

    @abc.abstractmethod
    async def get_paper(self, paperId: str) -> Optional[Paper]:
        """获取某篇论文的详情"""
        return None

    @abc.abstractmethod
    async def get_references(self, paper: Paper) -> AsyncIterable[Paper]:
        """获取某篇论文的引文"""
        return

    @abc.abstractmethod
    async def filter_papers(self, papers: AsyncIterable[Paper]) -> AsyncIterable[Paper]:
        """
        在收集信息时过滤`Paper`，不会对被此方法过滤掉的`Paper`进行信息收集
        等同于dblp-crawler里的filter_publications_at_crawler
        """
        async for paper in papers:
            yield paper

    async def init_paper(self, paperId) -> Tuple[int, int]:
        init, refs = 0, 0
        if paperId not in self.papers or not self.papers[paperId]:
            paper = await self.get_paper(paperId)
            if not paper:
                return init, refs
            paperId = paper.paperId()
            init += 1
        else:
            paper = self.papers[paperId]
            paperId = paper.paperId()
        self.papers[paperId] = paper
        if paperId in self.checked:
            return init, refs
        async for new_paper in self.filter_papers(self.get_references(paper)):
            if not new_paper:
                continue
            new_paperId = new_paper.paperId()
            if paperId not in self.ref_idx:
                self.ref_idx[paperId] = set()
            self.ref_idx[paperId].add(new_paperId)
            if new_paperId not in self.papers or not self.papers[new_paperId]:
                self.papers[new_paperId] = new_paper
                refs += 1
        self.checked.add(paperId)
        return init, refs

    async def bfs_once(self) -> int:
        tasks = [self.init_paper(paperId) for paperId in list(self.papers.keys())]
        total_init, total_refs = 0, 0
        for init, refs in await asyncio.gather(*tasks):
            total_init += init
            total_refs += refs
        logger.info("Initializing %s papers and %s refernces" % (total_init, total_refs))
        return total_init, total_refs


class Summarizer(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    async def filter_papers(self, papers: AsyncIterable[Paper]) -> AsyncIterable[Paper]:
        """
        在输出时过滤`Paper`，被过滤掉的`Paper`将不会出现在输出中
        等同于dblp-crawler里的filter_publications_at_output
        """
        async for paper in papers:
            yield paper

    @abc.abstractmethod
    async def write_paper(self, paper: Paper) -> None:
        pass

    @abc.abstractmethod
    async def write_reference(self, paper: Paper, reference: Paper) -> None:
        pass

    async def write(self, crawler: Crawler) -> None:
        exist_papers = set()
        async def wrapper(iter: Iterable[Paper]):
            for i in iter:
                yield i
        async for paper in self.filter_papers(wrapper(crawler.papers.values())):
            await self.write_paper(paper)
            exist_papers.add(paper.paperId())
        for paperId, refs_paperId in crawler.ref_idx.items():
            if paperId not in exist_papers:
                continue
            for ref_paperId in refs_paperId:
                if refs_paperId not in exist_papers:
                    continue
                await self.write_reference(crawler.papers[paperId], crawler.papers[ref_paperId])

    def __call__(self, crawler: Crawler) -> Any:
        return self.write(crawler)



from typing import Optional, Dict
import os
from datetime import datetime, timedelta
import json

import aiohttp
import logging
from aiofile import async_open
from asyncio import Semaphore

logger = logging.getLogger("common")

http_sem = Semaphore(8)
file_sem = Semaphore(512)


def getenv_int(key) -> int:
    cache_days = os.getenv(key)
    if cache_days is not None:
        try:
            return int(cache_days)
        except:
            pass
    return None


async def fetch_item(url: str) -> Optional[Dict]:
    async with http_sem:
        try:
            async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(verify_ssl=False)) as session:
                async with session.get(url) as response:
                    logger.info("fetch url: %s" % url)
                    text = await response.text()
                    try:
                        return json.loads(text)
                    except Exception as e:
                        logger.error(" json err: %s: %s" % (e, text))
        except Exception as e:
            logger.error("fetch err: %s" % e)
    return None


def get_cache_datetime(path) -> datetime:
    return datetime.fromtimestamp(os.path.getmtime(path))


async def download_item(url: str, path: str, cache_days: int) -> Optional[Dict]:
    save_path = os.path.join("save", path)
    if os.path.isfile(save_path):
        if datetime.now() < get_cache_datetime(save_path) + timedelta(days=cache_days):
            async with file_sem:
                try:
                    async with async_open(save_path, 'r') as f:
                        logger.info("use cache: %s -> %s" % (path, url))
                        text = await f.read()
                        return json.loads(text)
                except:
                    logger.debug(" no cache: %s" % save_path)
        else:
            logger.info("old cache: %s" % save_path)

    async with http_sem:
        try:
            async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(verify_ssl=False)) as session:
                async with session.get(url) as response:
                    logger.info(" download: %s <- %s" % (path, url))
                    text = await response.text()
                    data = json.loads(text)
                    os.makedirs(os.path.dirname(save_path), exist_ok=True)
                    async with async_open(save_path, 'w') as f:
                        await f.write(text)
                    return data
        except Exception as e:
            logger.error(" down err: %s" % e)
    return None



import logging
import re
from citation_crawler import Summarizer, Paper

from neo4j import Session

'''Use with dblp-crawler'''

logger = logging.getLogger("graph")


def add_paper(tx, paper: Paper):
    dblp_id = paper.dblp_id()
    doi = paper.doi()
    if doi:
        if not re.match(r"^https*://doi.org/", doi):
            doi = "https://doi.org/" + doi
    if dblp_id and doi:
        tx.run("MERGE (p:Publication {key: $key}) "
               "ON CREATE SET p.title=$title, p.year=$year, p.doi=$doi",
               key=dblp_id,
               title=paper.title(),
               year=paper.year(),
               doi=doi)
    elif dblp_id:
        tx.run("MERGE (p:Publication {key: $key}) "
               "ON CREATE SET p.title=$title, p.year=$year",
               key=dblp_id,
               title=paper.title(),
               year=paper.year())
    elif doi:
        tx.run("MERGE (p:Publication {doi: $doi}) "
               "ON CREATE SET p.title=$title, p.year=$year",
               doi=doi,
               title=paper.title(),
               year=paper.year())


def match_statement(p: Paper):
    dblp_id = p.dblp_id()
    if dblp_id:
        return "MATCH ({n}:Publication {{key: ${n}}}) ", dblp_id
    doi = p.doi()
    if doi:
        if not re.match(r"^https*://doi.org/", doi):
            doi = "https://doi.org/" + doi
        return "MATCH ({n}:Publication {{doi: ${n}}}) ", doi
    return None, None


def add_reference(tx, a: Paper, b: Paper):
    sta, vla = match_statement(a)
    stb, vlb = match_statement(b)
    if sta is None or stb is None:
        return
    statement = sta.format(n='a') + stb.format(n='b') + "MERGE (a)-[:CITE]->(b)"
    tx.run(statement, a=vla, b=vlb)


class Neo4jSummarizer(Summarizer):
    def __init__(self, session: Session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session

    async def write_paper(self, paper) -> None:
        self.session.execute_write(add_paper, paper)

    async def write_reference(self, paper, reference) -> None:
        self.session.execute_write(add_reference, paper, reference)



import argparse
import asyncio
import logging

from dblp_crawler.keyword.arg import add_argument as add_argument_kw, parse_args as parse_args_kw
from citation_crawler.arg import add_argument_pid, parse_args_pid
from citation_crawler.crawlers import SemanticScholarCrawler
from citation_crawler.summarizers import NetworkxSummarizer, Neo4jSummarizer

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('citation_crawler')

parser = argparse.ArgumentParser()

parser.add_argument("-y", "--year", type=int, help="Only crawl the paper after the specified year.", default=2000)
add_argument_kw(parser)
add_argument_pid(parser)


def func_parser(parser):
    args = parser.parse_args()
    year = args.year
    keywords = parse_args_kw(parser)
    pid_list = parse_args_pid(parser)
    logger.info(f"Specified keyword rules: {keywords.rules}")
    logger.info(f"Specified paperId list for init: {pid_list}")
    return year, keywords, pid_list


async def filter_papers_at_crawler(papers, year, keywords):
    async for paper in papers:
        if (paper.year() is None or paper.year() >= year) and keywords.match_words(paper.title()):
            yield paper


async def filter_papers_at_output(papers, year, keywords):
    async for paper in papers:
        if (paper.year() is None or paper.year() >= year) and keywords.match(paper.title()):
            yield paper


async def bfs_to_end(graph, limit: int = 0):
    while min(*(await graph.bfs_once())) > 0 and (limit != 0):
        logger.info("Still running......")
        limit -= 1


subparsers = parser.add_subparsers(help='sub-command help')


# --------- for SemanticScholarCrawler ---------

class DefaultSemanticScholarCrawler(SemanticScholarCrawler):
    def __init__(self, year, keywords, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.year = year
        self.keywords = keywords

    async def filter_papers(self, papers):
        """在收集信息时过滤`Paper`，不会对被此方法过滤掉的`Paper`进行信息收集"""
        async for paper in filter_papers_at_crawler(papers, self.year, self.keywords):
            yield paper


# --------- for NetworkxGraph ---------

class DefaultNetworkxSummarizer(NetworkxSummarizer):
    def __init__(self, year, keywords, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.year = year
        self.keywords = keywords

    async def filter_papers(self, papers):
        """在输出时过滤`Paper`，被过滤掉的`Paper`将不会出现在输出中"""
        async for paper in filter_papers_at_output(papers, self.year, self.keywords):
            yield paper


parser_nx = subparsers.add_parser('networkx', help='Write results to a json file.')
parser_nx.add_argument("--dest", type=str, required=True, help=f'Path to write results.')


def func_parser_nx(parser):
    year, keywords, pid_list = func_parser(parser)
    args = parser.parse_args()
    dest = args.dest
    logger.info(f"Specified dest: {dest}")
    crawler = DefaultSemanticScholarCrawler(year, keywords, pid_list)
    asyncio.get_event_loop().run_until_complete(bfs_to_end(crawler))
    summarizer = DefaultNetworkxSummarizer(year, keywords)
    asyncio.get_event_loop().run_until_complete(summarizer(crawler))
    asyncio.get_event_loop().run_until_complete(summarizer.save(dest))


parser_nx.set_defaults(func=func_parser_nx)


# --------- for Neo4jGraph ---------


class DefaultNeo4jSummarizer(Neo4jSummarizer):
    def __init__(self, year, keywords, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.year = year
        self.keywords = keywords

    async def filter_papers(self, papers):
        """在输出时过滤`Paper`，被过滤掉的`Paper`将不会出现在输出中"""
        async for paper in filter_papers_at_output(papers, self.year, self.keywords):
            yield paper


# --------- Run ---------
args = parser.parse_args()
args.func(parser)



import argparse
import importlib


def add_argument_pid(parser: argparse.ArgumentParser, *flags, dest: str = 'pid') -> None:
    if len(flags) <= 0:
        flags = ['-p', '--pid']
    parser.add_argument(
        *flags, dest=dest, action='append', required=False, default=[],
        help=f'Specified a list of paperId to start crawling.'
    )


def parse_args_pid(parser: argparse.ArgumentParser, pid_dest: str = 'pid'):
    args = parser.parse_args()
    pid_list = []
    for pid_s in args.__getattribute__(pid_dest):
        try:
            pid = eval(pid_s)
            if isinstance(pid, str):
                pid_list.append(pid)
            else:
                pid_list.extend(pid)
        except:
            pid_list.append(pid_s)
    return pid_list



import abc
from typing import Optional, Iterable


class Author(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def authorId(self) -> str:
        return ''

    @abc.abstractmethod
    def name(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def dblp_pid(self) -> Optional[str]:
        return None

    def __dict__(self) -> dict:
        d = {}
        if self.authorId():
            d['authorId'] = self.authorId()
        if self.name():
            d['name'] = self.name()
        if self.dblp_pid():
            d['dblp_pid'] = self.dblp_pid()
        return d


class Paper(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def paperId(self) -> str:
        return ''

    @abc.abstractmethod
    def dblp_id(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def title(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def year(self) -> Optional[int]:
        return None

    @abc.abstractmethod
    def doi(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    async def authors(self) -> Iterable[Author]:
        return

    async def __dict__(self) -> dict:
        d = {}
        if self.paperId():
            d['paperId'] = self.paperId()
        if self.dblp_id():
            d['dblp_id'] = self.dblp_id()
        if self.title():
            d['title'] = self.title()
        if self.year():
            d['year'] = self.year()
        if self.doi():
            d['doi'] = self.doi()
        d['authors'] = []
        async for author in self.authors():
            d['authors'].append(author.__dict__())
        return d
